#include "RACE_LogfileOffset.h"
#include "RACE_Accumulator.h"
#include "errorhandling/envdebug.h"
#include "errorhandling/errorhandling.h"

static envdebug *debug = new envdebug("RACE_Accumulator");

RACE_AccumulatorObserver::~RACE_AccumulatorObserver()
{
}

RACE_Accumulator::RACE_Accumulator(RACE_WriterInterface *writer,size_t size,size_t blocksize):m_Writer(writer)
{
  mp_entry = new SEntry[size];
  RACE_COND_FAIL(mp_entry==NULL,"%s:%d: RACE_Accumulator: No more Memory",__FILE__,__LINE__);
  m_io = new iovec[size];
  RACE_COND_FAIL(m_io==NULL,"%s:%d: RACE_Accumulator: No more Memory",__FILE__,__LINE__);
  for (size_t i = 0; i < size; i++) {
    m_io[i].iov_base = NULL;
    m_io[i].iov_len = 0;
  }
  m_CurrentOffset = 0;
  m_CurrentSize = 0;
  m_Index=0;
  m_IndexMax=size;
  m_BlockSize=blocksize;
  m_WriteSize=blocksize;
  m_written=0;
  m_AccuObserver=0;
  m_inflush=0;
}

RACE_Accumulator::~RACE_Accumulator()
{
  delete[] mp_entry;
  delete[] m_io;
}

void RACE_Accumulator::Add(RACE_Dumpable *d)
{
  if (d == NULL) {
    RACE_ERROR ("RACE_Accumulator::Add: argument d is NULL");
    return; // we cannot add any fragment from non existing dumpable
  }
  if (d->IsFragmented ()) {
    debug->print ("RACE_Accumulator::Add: is_fragmented");
    bool first = true;
    for (d->begin (); ! d->is_done (); d->advance ()) {
      debug->print ("RACE_Accumulator::Add: fragment %p length %d/%d", d, d->length (), d->total_length ());
      AddFrag (d, d->is_last (), first); // fragmented are only notified once
      first = false;
    }
  }
  else {
    // Optimization for the common non fragmented case
    AddFrag (d, true, true);
  }
  d->release ();
}

void RACE_Accumulator::AddFrag (RACE_Dumpable *d, bool notify_completion, bool notify_offset)
{
  debug->print("RACE_Accumulator::AddFrag: %p currentsize=%d add %d", d, m_CurrentSize, d->length ());
  if (m_Index >= m_IndexMax) {
    Flush();
  }
  if (mp_entry[m_Index].mp_dumpable) {
    RACE_ERROR ("RACE_Accumulator::AddFrag: dangling dumpable detected with index %i", m_Index);
    Release_Dumpable (m_Index);
  }
  mp_entry[m_Index].mp_dumpable = ACE_dynamic_cast(RACE_Dumpable*,d->add_ref());
  if (notify_offset) 
  {
    // notify the current offset to the application
    mp_entry[m_Index].mp_dumpable->InformOffset(RACE_LogfileOffset(m_Writer->PartNo(), m_CurrentOffset));
  }
  mp_entry[m_Index].m_notify = notify_completion;
  m_io[m_Index].iov_base = d->rd_ptr();
  m_io[m_Index].iov_len = d->length();
  m_CurrentOffset += d->length();

  // the while loop is (not entered)/(terminates) if the accumulator has enough space to store the dumpable
  // otherwise we have to manage fragmentation and flushing
  while (m_CurrentSize + mp_entry[m_Index].mp_dumpable->remaining() >= m_WriteSize)
  {
    m_io[m_Index].iov_len = m_WriteSize - m_CurrentSize;
    mp_entry[m_Index].mp_dumpable->rd_ptr(m_io[m_Index].iov_len);

    // write the first part according to the given iov_len
    (void)WriteVec(m_io, m_Index + 1);
       
    size_t remaining_bytes = mp_entry[m_Index].mp_dumpable->remaining();
    if (remaining_bytes > 0) 
    {
      // have to process "the rest of it"
      debug->print("RACE_Accumulator::AddFrag: remaining_bytes=%d", remaining_bytes);

      // ... keep the dumpable pointer in a safe variable
      {
        SEntry temp_entry = mp_entry[m_Index];
        mp_entry[m_Index].mp_dumpable = NULL; // this prevents the temp_entry.mp_dumpable from getting deleted
        FlushAfterWrite (); // this resets m_Index to 0
        mp_entry[m_Index] = temp_entry;
      }
      m_io[m_Index].iov_base = mp_entry[m_Index].mp_dumpable->rd_ptr ();
      m_io[m_Index].iov_len = remaining_bytes;
    }
    else 
    {
      m_Index++; // the current index has to be processed as well
      FlushAfterWrite (); // this resets m_Index to 0
      return;
    }
  }

  m_CurrentSize += m_io[m_Index].iov_len; // adjust current used size within the accumulator
  m_Index++; // this marks the next available space within m_io that is used by the following invocation of AddFrag
}

int RACE_Accumulator::WriteVec(const  struct  iovec  *  vector,  int count)
{
    int ret=m_Writer->WriteV(vector,count);
    m_written+=ret;
    int frag=(m_written&(m_BlockSize-1)); // m_written%BlockSize
    m_WriteSize=m_BlockSize-frag;
    debug->print ("RACE_Accumulator::WriteVec: wrote %d bytes WriteSize=%d",
		  ret,m_WriteSize);
    return ret;
}

void RACE_Accumulator::Release_Dumpable (int i)
{
  if (mp_entry[i].mp_dumpable) {
    if (mp_entry[i].m_notify) {
      mp_entry[i].mp_dumpable->InformCompleted ();
    }
    mp_entry[i].mp_dumpable->release ();
    mp_entry[i].mp_dumpable = NULL;
  }
}

void RACE_Accumulator::FlushAfterWrite()
{
  if(m_AccuObserver) {
    m_AccuObserver->NotifyFlush();
  }
  for(int i=0;i<m_Index;i++) {
    Release_Dumpable (i);
  }
  m_Index=0;
  m_CurrentSize=0;
}

void RACE_Accumulator::Flush()
{
  size_t nw;
  m_inflush++;
  if(m_Index>0)
  {
    nw=WriteVec(m_io,m_Index);
    debug->print("RACE_Accumulator::Flush(): %d Bytes written",nw);
    FlushAfterWrite();
  }
  else
  {
    debug->print("RACE_Accumulator::Flush(): was empty");
  }
  m_inflush--;
}

void RACE_Accumulator::LogSwitch()
{
  if(m_inflush==0)
  {
    Flush();
  }
  m_Writer->SwitchPart();
  m_CurrentOffset=0;
  m_written=0;
}
