#ifndef RACE_ACCUMULATOR_H
#define RACE_ACCUMULATOR_H
#include<errorhandling/errorclass.h>
#include"RACE_Dumpable.h"
#include<sys/uio.h>

//##Documentation
//##  Abstract Interface for Writing a file.
//##  The WriteV - Method uses the writev(2)-Systemcall to gather blocks
//##  togther without copiing

class RACE_WriterInterface
{
public:
  virtual ~RACE_WriterInterface () {}
    //##Documentation
    //## Interface like the Systemcall writev(2)
    //## if an error occurs, a exception will thrown
  virtual int WriteV(const  struct  iovec  *  vector,  int count) throw (Systemerror)=0;
  virtual void Close()=0;
  virtual void Open() throw (Systemerror)=0;
    //##Documentation
    //## Returns the number of the current part of the logfile
  virtual int PartNo()=0;
    //##Documentation
    //## Switch the Part
  virtual void SwitchPart() throw (Systemerror)=0;
};

//##Documentation
//## Abstract Interface for observing the RACE_Accumulator
//## With this Interface a minimal flush rate can implemented
//## To avoid extra synchronisation, we want to send a FlushDumpable
//## But here we don't know what "Sending" is and we must decouple the
//## Implemantation
class RACE_AccumulatorObserver
{
public:
  virtual void NotifyFlush()=0;
  virtual ~RACE_AccumulatorObserver();
};

//##Documentation
//##  The Accumulator accumulates Dumpables to collect systemsize Blocks
//##  for instance  4K-Blocks

class RACE_Accumulator
{
public:
    //##Documentation
    //## blocksize must be a 2^n 
  RACE_Accumulator(RACE_WriterInterface *writer,size_t size=20,size_t blocksize=4096);
  ~RACE_Accumulator();
    //##Documentation
    //## add a Dumpable
  void Add(RACE_Dumpable *d);
    //##Documentation
    //## force the writing the collected blocks to the Writer
  void Flush();
    //##Documentation
    //## get the current offset in the current part
  size_t GetCurrentOffset() { return m_CurrentOffset; }
    //##Documentation
    //## reset the internal state for the switching the logfile
  void LogSwitch();
  void SetAccuObserver(RACE_AccumulatorObserver *obs) { m_AccuObserver=obs; }
protected:
  void FlushAfterWrite();
  void AddFrag(RACE_Dumpable *d, bool notify_completion, bool notify_offset);
  //##Documentation
  //## calls the Writer->WriteV and calculates m_WriteSize
  int WriteVec(const  struct  iovec  *  vector,  int count);
private:
  void Release_Dumpable (int i);

  struct SEntry
  {
    SEntry () 
      : mp_dumpable(NULL),
        m_notify(true)
      {}

    RACE_Dumpable *mp_dumpable;
        //##Documentation
        //## whether the dumpable is to be notified after write
    bool m_notify;
  };
  SEntry        *mp_entry;
    //##Documentation
    //## an POSIX-IO-vector see writev(2)
  struct iovec *m_io;

    //##Documentation
    //## The offset in the part
  off_t m_CurrentOffset;
    //##Documentation
    //## the current size of the block collected now
  size_t m_CurrentSize;
  int m_Index;
  int m_IndexMax;
    //##Documentation
    //## The size of the blocks to be written to the Writer
  size_t m_BlockSize;
  //##Documentation
  //## The size which should be written, normaly m_Blocksize. May be less
  //## in case of Snapshots and so on. It is determined by WriteVec
  size_t m_WriteSize;
  //##Documentation
  //## bytes written in file - state for WriteVec
  size_t m_written;
  RACE_WriterInterface *m_Writer;

  RACE_AccumulatorObserver *m_AccuObserver;
  //##Documentation
  //## avoid recursive calling of flush
  //## The InformOffsets may trigger an Flush indirectly
  int m_inflush;
};

#endif
