#include "RACE_LogIndexWriter.h"
#include <unistd.h> // write(2);
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cerrno>
#include <cstdio>
#include "errorhandling/errorhandling.h"
#include "errorhandling/envdebug.h"

static envdebug *debug = new envdebug ("RACE_LogIndexWriter");

RACE_LogIndexWriter::~RACE_LogIndexWriter()
{
  if(m_fdi!=-1)
  {
    Close();
  }
  if (m_indexname) {
    delete m_indexname;
    m_indexname = NULL;
  }
}

void RACE_LogIndexWriter::Open(const char *basefilename)
{
  set_indexname (basefilename);
  const char* indexfilename = m_indexname->get_fullname ();
  m_fdi=::open(indexfilename,O_CREAT|O_WRONLY,0644);
  RACE_COND_FAIL(m_fdi < 0, "RACE_LogIndexWriter::Open: can't open %s:%s", indexfilename, strerror(errno));
  debug->print("RACE_LogIndexWriter::Open: opened file %s", indexfilename);
}

void RACE_LogIndexWriter::SaveOldIndex(const char *basefilename,int version)
{
  set_indexname (basefilename);
  const char* indexfilename = m_indexname->get_fullname ();
  char* versionedindexfilename = m_indexname->get_versioned_fullname (version);
  int r = rename(indexfilename, versionedindexfilename);
  if (r < 0) {
    RACE_ERROR("rename(%s,%s): %s", indexfilename, versionedindexfilename, strerror(errno));
  }
  debug->print("RACE_LogIndexWriter::SaveOldIndex: renamed file %s to %s",
               indexfilename, versionedindexfilename);
  delete [] versionedindexfilename;
}


void RACE_LogIndexWriter::RemoveVersion(const char *basefilename,int version)
{
  set_indexname (basefilename);
  char* versionedindexfilename = m_indexname->get_versioned_fullname (version);
  unlink(versionedindexfilename);
  debug->print("RACE_LogIndexWriter::Remove: removed file %s", versionedindexfilename);
  delete [] versionedindexfilename;
}

void RACE_LogIndexWriter::WriteEntry(time_t unixtime,RACE_LogfileOffset offset)
{
  RACE_COND_FAIL(m_indexname == NULL, "RACE_LogIndexWriter::WriteEntry: got never opened");
  RACE_LogIndexLayout lie;
  lie.unixtime = unixtime;
  lie.offset   = offset;
  size_t n=::write(m_fdi,&lie,sizeof(RACE_LogIndexLayout));
  RACE_COND_FAIL(n != sizeof(RACE_LogIndexLayout),
                    "RACE_LogIndexWriter::WriteEntry: %d instead of %d written - Index of %s corrupted",
                    n, sizeof(RACE_LogIndexLayout), m_indexname->get_fullname ());
  debug->print("RACE_LogIndexWriter::WriteEntry: part=%d, offset=%d", offset.part, offset.offset);
}

void RACE_LogIndexWriter::Close()
{
  RACE_COND_FAIL(m_indexname == NULL, "RACE_LogIndexWriter::Close: got never opened");
  debug->print("RACE_LogIndexWriter::Close: closing file %s", m_indexname->get_fullname ());
  ::close(m_fdi);
  m_fdi=-1;
}

void RACE_LogIndexWriter::Remove()
{
  RACE_COND_FAIL(m_indexname == NULL, "RACE_LogIndexWriter::Remove: got never opened");
  const char* indexfilename = m_indexname->get_fullname ();
  debug->print("RACE_LogIndexWriter::Remove: removing file %s", indexfilename);
  unlink(indexfilename);
}

void RACE_LogIndexWriter::set_indexname (const char* basename)
{
  if (m_indexname == NULL) {
    m_indexname = new RACE_IndexFileName (basename);
    RACE_COND_FAIL (m_indexname == NULL, "%s:%d: memory allocation failure", __FILE__, __LINE__);
  }
  else
    RACE_COND_FAIL (strcmp (basename, m_indexname->get_basename ()) != 0,
                       "RACE_LogIndexWriter::set_indexname: mismatch in basename - old usage was %s and "
                       "new usage is %s", m_indexname->get_basename (), basename);
}
