#include "RACE_Logger.h"
#include "RACE_CmdDumpable.h"
#include "RACE_LogSwitchingStrategy.h"

envdebug *RACE_Logger::ms_debug = new envdebug("RACE_Logger");

/*------------- The Logger itself ----------------*/

RACE_Logger::RACE_Logger(RACE_WriterInterface *logfile,RACE_LoggerStrategy *strat, RACE_LogSwitchingStrategy *lswstat)
    : RACE_Synch_Task(strat,0),
      mp_logfile(logfile),
      m_accu(logfile),
      mp_LogSwitchingStrategy(lswstat),
      m_canceled(false),
      m_strat(strat),
      m_refcount(1)
{
  if(mp_LogSwitchingStrategy==0)
  {
    mp_LogSwitchingStrategy=new RACE_DefaultLogSwitchingStrategy(this);
    RACE_MEMORY_FAIL(mp_LogSwitchingStrategy==0,__FILE__,__LINE__,__FUNCTION__);
  }
  else
  {
      // the Strategy is created before the logger and needs to be informed
      mp_LogSwitchingStrategy->SetSwitcher(this);
  }
}

const char *RACE_Logger::GetName()
{
  return m_strat->GetQueueName();
}

void RACE_Logger::Log(RACE_Dumpable *dumpable)
{
  /*
   * During the Canceling-Operation in the Flush-Operation the Logging of
   * Snapshots may be triggered - this happens in the Loggerthread
   */
  if(m_canceled && dumpable->length() > 0 && !m_strat->IsCallerLoggerThread())
  {
    RACE_WARN("RACE_Logger::log: Logger canceled - discarding dumpable");
    dumpable->Discard();
    dumpable->release();
    return;
  }
  int r=msg_queue_.enqueue(dumpable);
  ms_debug->print("RACE_Logger::log enqueue returns %d",r);
  if(r<0 && errno==EWOULDBLOCK)
  {
    RACE_ERROR("EnQueuing TIMEOUT - Canceling");
    Cancel();
    // Cancel Imediately,
    // don't wait until the Canceldumpable reaches the logger
    m_canceled=true;
    return;
  }
  RACE_COND_FAIL(r<0,"Logger::log enqueue");
}

int RACE_Logger::svc(void)
{
  int count=1;
  RACE_Message_Block *mb;
  long periode_in_s = m_strat->periode_in_s();
  time_t nextperiodetime = time_t(NULL)+periode_in_s;
  time_t now;

  try
  {
    m_strat->SetLoggerThreadId(ACE_OS::thr_self());
    mp_logfile->Open(); // At start of the Thread open the Logfile is not done
    for(;!(m_canceled && msg_queue_.is_empty()); count++)
    {
	if(periode_in_s > 0 && (now=time(NULL))>nextperiodetime)
	{
	    m_strat->periodic_callback(now);
	    nextperiodetime = now+periode_in_s;
	}
	ms_debug->print("in iteration %d before getq()",count);
	if (this->getq(mb) == -1)
	{
	    bool cont = (errno==EWOULDBLOCK); // errno may overwritten by flush
	    m_accu.Flush(); // If error or timeout , we want to have all on disk
	    if(cont) continue;
	    RACE_ERROR("in iteration %d, got result -1, exiting\n",count);
	    break;
	}
	RACE_Dumpable *d=ACE_dynamic_cast(RACE_Dumpable *,mb);
	int length = d->total_length();

	if (length > 0)
	{
	    ms_debug->print("in iteration %d, length = %d fragmented=%s\n",
			    count,length,(d->IsFragmented() ? "true" : "false"));
	    if(d->IsDiscardable() && m_strat->is_critical_watermark_reached())
	    {
		RACE_WARN("RACE_Logger: Discarding dispensable Dumpable");
		d->Discard();
	    }
	    else
	    {
		m_accu.Add(d);
		mp_LogSwitchingStrategy->InfoCurrentOffset(m_accu.GetCurrentOffset());
	    }
	}
	else
	{
	    if(ACE_dynamic_cast(RACE_CancelCmdDumpable *,d))
	    {
		ms_debug->print("RACE_Logger: RACE_CancelCmdDumpable");
		m_accu.Flush(); // some Dumpables may produced here
		m_canceled = true; // dissallow queueing
	    }
	    else if(ACE_dynamic_cast(RACE_FlushCmdDumpable *,d))
	    {
		ms_debug->print("RACE_Logger: RACE_FlushCmdDumpable");
		m_accu.Flush();
	    }
	    else
	    {
		RACE_ERROR("RACE_Logger received an Dumpable with length=%d, queuelength=%d",length,msg_queue_.message_count());
	    }
	    d->release();
	}
    }
    m_accu.Flush(); // Flush the last Dumpables
    m_strat->logger_terminated(RACE_LoggerStrategy::TC_NORMAL);
    mp_logfile->Close();
  }
  catch(std::exception &ex)
  {
      m_canceled=true; // Don't accept any more Dumpables 
      m_strat->logger_terminated(RACE_LoggerStrategy::TC_EXCEPTION);
      RACE_ERROR("RACE_Logger terminated: %s ",ex.what());
  }
  return 0;
}



void RACE_Logger::Cancel()
{
  RACE_Dumpable *d = new RACE_CancelCmdDumpable();
  /*
   * In all cases, the CancesDumbable must be enqueued
   * The CActiveObject must be possible to flush its queue and so on
   */
  int r=msg_queue_.enqueue_force(d);
  ms_debug->print("RACE_Logger::Cancel: enqueue_force returns %d",r);
  RACE_COND_FAIL(r<0,"CActiveConsumer::enqueue_force");
}

void RACE_Logger::Flush()
{
  RACE_Dumpable *d = new RACE_FlushCmdDumpable();
  Log(d);
}

RACE_Logger* RACE_Logger::Duplicate()
{
  m_refcount++;
  ms_debug->print("RACE_Logger::duplicate: %d References",m_refcount);
  return this;
}

void RACE_Logger::Release()
{
  ms_debug->print("RACE_Logger::release: %d ",m_refcount);
  m_refcount--;
  if(m_refcount == 0)
  {
    ms_debug->print("RACE_Logger::release: would delete myself: thr_count=%d",
                 thr_count());
  }
}

void RACE_Logger::SwitchLog()
{
  m_accu.LogSwitch();
}

void RACE_Logger::InfoSnapshotWritten()
{
  mp_LogSwitchingStrategy->InfoSnapshotWritten();
}

RACE_Logger::~RACE_Logger()
{
  ms_debug->print("RACE_Logger::~RACE_Logger()");
}


#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)
template class RACE_Message_Queue<ACE_MT_SYNCH>;
template class RACE_Task<ACE_MT_SYNCH>;
#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)
#pragma instantiate RACE_Message_Queue<ACE_MT_SYNCH>;
#pragma instantiate RACE_Task<ACE_MT_SYNCH>;
#endif

