#ifndef RACE_MESSAGE_BLOCK
#define RACE_MESSAGE_BLOCK

#include<sys/types.h>
#include<ace/Synch_T.h>

#include"RACE_Data_Block.h"

//##Documentation
//## All Visitors/Semantics inherit from this interface
class RACE_MessageVisitor
{
  public:
  //##Documentation
  //## make artificialy an interface
  virtual const char *name()=0;
};


//##Documentation
//##  An RACE_Message_Block is modeled after the message data
//##  structures used in System V STREAMS.  Its purpose is to
//##  enable efficient manipulation of arbitrarily large messages
//##  without incurring much memory copying overhead.  Here are the
//##  main characteristics of an ACE_Message_Block:
//##  - Contains a pointer to a reference-counted
//##    RACE_Data_Block, which in turn points to the actual data
//##    buffer.  This allows very flexible and efficient sharing of
//##    data by multiple RACE_Message_Block objects.
//##  - It contains one or a chain of RACE_Data_Blocks (fragment-chain)
//##  - RACE_Message_Blocks can be linked together in a doubly linked fashion
//##    to form a queue of messages (this is how RACE_Message_Queue works).
//##  (stolen from ACE_Message_Block)
//##  Further, the RACE_Message_Block are based on following ideas:
//##  * The RACE_Data_Blocks contains raw data read from or to write to disk
//##    or from an (network-) device. 
//##  * The RACE_Message_Block can be derived. A derivation can have 
//##    accessor methods to the data in the RACE_Data_Blocks
//##  * By walking up a protocol stack, the RACE_Message_Block may
//##    be replaced by an more specific derivation or switched
//##    to an special (see RACE_Dumpable for writing to an Logfile)
//##  * Logical seen, it contains only the data and abstract from where
//##    it is read and where it will to be go - just the data
//##  * for further interpretion, the abstract method callvisitor calls an 
//##    abstract visitor. Problem specific subsclasses have to refine it.
//##    (Semantic - or Visitor-Pattern)
class RACE_Message_Block
{
 public:
    //##Documentation
    //## generates a empty Message_Block with size len
    RACE_Message_Block (size_t len);

    //##Documentation
    //## Generates a Messageblock with a existent Data_Block
    //## Important for the read side
    //## frag - indicates if the datablock is fragmented
    RACE_Message_Block(RACE_Data_Block *db,bool frag);

    // = Operations on Message Data
    //##Documentation
    //## Resets the Message Block data to contain nothing, i.e., sets the
    //## read and write pointers to align with the base.
    void reset ();

    //##Documentation
    //## Get message data.
    char *base () const;

    //##Documentation
    //## Return a pointer to 1 past the end of the allocated data in a message.
    char *end () const;

    //##Documentation
    //## pointer to the base of the first data block
    char *first_base () const;

    //##Documentation
    //##  Copies <n> bytes from <buf> into the Message_Block starting at
    //##  the <wr_ptr> offset.  Return 0 and increment <wr_ptr> by <n> if
    //##  the method succeeds.  Returns -1 if the size of the message is
    //##  too small, i.e., for this to work correct, <end> must be >=
    //##  <wr_ptr>.
    int copy (const char *buf, size_t n) throw (BufferOverflow);

    //##Documentation
    //##  Get the read pointer.
    char *rd_ptr () const;

    //##Documentation
    //##  Set the read pointer to <ptr>.
    void rd_ptr (char *ptr);

    //##Documentation
    //##  Set the read pointer ahead <n> bytes.
    //##  and return the position befor setting ahead
    char *rd_ptr (size_t n);
    
    //##Documentation
    //## relative position of the readpointer
    size_t rd_rel() { return rd_ptr_; }

    //##Documentation
    //## relative position of the writepointer
    size_t wr_rel() { return wr_ptr_; }

    //##Documentation
    //##  Get the write pointer.
    char *wr_ptr () const;

    //##Documentation
    //##  Set the write pointer to <ptr>.
    char* wr_ptr (char *ptr) throw (BufferOverflow);

    //##Documentation
    //##  Set the write pointer ahead <n> bytes.
    //##  and returns the old position.
    //##  It throws BufferOverflow if the fragment is full.
    //##  allows writing compactly things like memset(wr_ptr(10),10,c);
    char* wr_ptr (size_t n) throw (BufferOverflow);
    
    //##Documentation
    //## aligns the wrptr to a boundary
    //## returns the new ptr
    char *wrptr_align(short potenz2);
    
    //##Documentation
    //## Get the number of bytes in the Message Block
    size_t size () const;

    //##Documentation
    //## Get the length of the message
    //## CAUTION: Different from ACE: this is the real length,
    //##          independent from the rd_ptr
    size_t length () const;

    //##Documentation
    //## remaining : the count of unconsumed bytes depending on the rd_ptr
    size_t remaining () const;
    
    //##Documentation
    //## Get the length of all DataBlocks in the <Message_Block>
    size_t total_length () const;
    
    //##Documentation
    //## Get the total number of bytes in all DataBlocks in the  <Message_Block>
    size_t total_size () const;
    
    //##Documentation
    //## returns the length of free memory in the current Datablock
    size_t freespace() { return db->freespace(); }

    //##Documentation
    //## adds a new datablock
    void addblock(RACE_Data_Block *block);

    //##Documentation
    //## allocates a new block for growing messageblocks
    void expand ();
    
    
    bool IsFragmented () const;

    //##Documentation
    //##iteration
    void begin ();

    bool is_done () const;

    bool is_last () const;

    void advance ();

  // = Operations on the whole Message
  //##Documentation
  //## Return a "fake" copy that increments our reference count by 1.
  RACE_Message_Block *add_ref ();

  //##Documentation
  //## Return a "shallow" copy.
  RACE_Message_Block *duplicate () const;

  //##Documentation
  //## return the datablock-chain
  //## the chain is duplicated
  RACE_Data_Block *duplicate_chain();

  //##Documentation
  //## Return a deep copy
  RACE_Message_Block *clone () const;

  //##Documentation
  //## Decrements our reference count by 1. If there is no reference present any more the
  //## destructor will be called.
  void release ();

  //##Documentation
  //## Get link to next message.
  RACE_Message_Block *next () const;

  //##Documentation
  //## Set link to next message.
  void next (RACE_Message_Block *n);

  // = Pointer to the <Message_Block> directly behind in the <ACE_Message_Queue>.

  //##Documentation
  //## Get link to prev message.
  RACE_Message_Block *prev () const;

  //##Documentation
  //## Set link to prev message.
  void prev (RACE_Message_Block *p);

  //##Documentation
  //## Hook for the Reading Semantik
  virtual void CallVisitor(RACE_MessageVisitor *visitor);

 private:
  //##Documentation
  //## Prohibited default operations.
  RACE_Message_Block (const RACE_Message_Block&);

  RACE_Message_Block& operator= (const RACE_Message_Block&);

  RACE_Data_Block *db; //the current Messageblock

  RACE_Data_Block *first_;

  size_t block_size_;

  size_t rd_ptr_; // Relative

  size_t wr_ptr_; // Relative

  ACE_Thread_Mutex _refcountmutex;

  unsigned _refcount;

 protected:
  RACE_Data_Block *firstblock();
  RACE_Data_Block *curblock();
  //##Documentation
  //## Protected destructor prevents from bypassing the reference count mechanism.
  virtual ~RACE_Message_Block ();

  //##Documentation
  //##  The cloning of derivations is only posible with a virtual clone
  //##  operation. It is called from duplicate ().
  //##  
  //## The derivation has to call a Constructor which initialize the 
  //## data for bookkepping purpose to invalid
  //## and copy the data helping to access to the copy. 
  virtual RACE_Message_Block *CloneManagementData () const;

  //##Documentation
  //##  The default constuctur is only for cloning the management data
  RACE_Message_Block ();

  //##Documentation
  //## set current data
  void reset_current ();

  RACE_Message_Block *next_;

  RACE_Message_Block *prev_;
};

#endif
