#ifndef RACESNAPSHOTTRIGGER_H
#define RACESNAPSHOTTRIGGER_H
#include<ace/Synch.h>

//##Documentation
//## A simple Queue for managing Snapshots
class RACE_SnapshotQueue
{
public:
  class Entry
  {
  public:
    Entry(time_t timestamp,int pending)
      :m_timestamp(timestamp),m_pending(pending),mp_next(0)
    {
    }
    virtual ~Entry();
    time_t GetTimestamp() { return m_timestamp; }
    int GetPending() { return m_pending; }
    int DecrPending() { return --m_pending; }
    virtual void ConnectionLog()=0;
  private:
    time_t m_timestamp;
    int  m_pending;
    Entry *mp_next;
    friend class ::RACE_SnapshotQueue;
  };
  RACE_SnapshotQueue():mp_first(0),mpp_last(&mp_first) {}
  virtual ~RACE_SnapshotQueue();
  void GetCountAndSum(int& count,int& sumpending);
  void enqueueSnapshotEntry(Entry *e);
  void dequeueSnapshotEntry();
  bool IsEmpty() { return mp_first==0; }
  time_t GetTopTimestamp() { return mp_first->GetTimestamp(); }
  int    GetTopPending()   { return mp_first->GetPending(); }
  int    DecrTopPending()  { return mp_first->DecrPending(); }

  Entry *GetTop() { return mp_first; }

  //##Documentation
  //## gets the entry with timestamp from the queue
  //## Returns NULL if no entry with exactly matching Timestamp
  Entry *GetEntry(time_t timestamp);
protected:
  Entry *GenNewEntry(time_t timestamp,int pending);
private:
  Entry *mp_first;
  Entry **mpp_last;
};

const int ST_UNKOWN=-1;

//##Documentation
//##  Interface for creation and Collecting Snapshots,
//##  used by RACE_SnapshotTrigger
class RACE_SnapshotCreator
{
public:
  //##Documentation
  //##  has to return the total number of Dumpables pending
  //##  returns ST_UNKOWN if not known
  //##  The default-Implemantation does not know and returns ST_UNKOWN.
  //##  For example, for OnEventSnapshots it is known.
  virtual int GetPendingTotal();
  //##Documentation
  //## has to return the number of Dumpables pending for the current
  //## Snapshot
  //## returns ST_UNKOWN if not known
  //##  The default-Implementation does not know and returns ST_UNKOWN.
  //## In WholeSnapshots it is known
  virtual int GetPendingForYou();
  virtual RACE_SnapshotQueue::Entry *GenNewEntry(time_t timestamp,int pending)=0;
  virtual ~RACE_SnapshotCreator();
};

//##Documentation
//##  Triggers the logging of a snapshot if all pending Dumpables
//##  are written.
//##  Thus the not written Dumpables are referenced in a Snapshot
//##  This is a strategy-Class
//##
//##  PrepareSnapshot() is called from the Main-Thread,
//##  InformOffset/InformCompleted is called from the Loggerthread, thus we need Koordination
class RACE_SnapshotTrigger
{
public:
  RACE_SnapshotTrigger(const char *description,RACE_SnapshotQueue *sq=NULL);
  virtual ~RACE_SnapshotTrigger();

  void PrepareSnapshot(RACE_SnapshotCreator* snapcreator, time_t timestamp);

  bool IsSnapshotPending() { return !m_queue->IsEmpty(); }
  RACE_SnapshotQueue::Entry *GetCurrent()     { return m_queue->GetTop(); }

  //##Documentation
  //##  This is called in the InformOffset/-Completed methods when the Dumpable
  //##  is written or discarded
  //##  In the view of the Snapshottrigger, there is no difference
  //##  between Inserting and Discarding
  void InformInsert();

  //##Documentation
  //##  This is called in the InformOffset/-Completed methods when the Dumpable
  //##  is written
  //##  it allows a Insertion overtaking an other in the case of jumbling
  void InformInsert(time_t timestamp);

  //##Documentation
  //## Getting the Entry from the Queue
  //## Returns 0 if not found
  RACE_SnapshotQueue::Entry *GetSnapshot(time_t timestamp) { return m_queue->GetEntry(timestamp); }
private:

  //##Documentation
  //## This is a helper used by both InformInsert methods.
  //## If the next pending snapshots don't wait for pending Dumpables, they will have to be logged
  //## as well. If a later snapshot is ready before an earlier, it will be triggered
  //## after its decendant is flushed.
  void InformInsertConclusion();

  ACE_Thread_Mutex m_mutex;

  //##Documentation
  //## Description for Debugging
  const char *m_description;

  //##Documentation
  //## last timestamp for controlling rthe monotonie of the prepare - Calls
  time_t m_lastTimestamp;

  RACE_SnapshotQueue *m_queue;
};
#endif

