#include "RACE_Thread.h"
#include "errorhandling/envdebug.h"

#include <ace/Thread.h>

envdebug *RACE_Thread::debug = new envdebug("RACE_Thread");

RACE_Thread::RACE_Thread(const char *threadname)
    :state_changed_(mutex_),
     name_(threadname ? strdup(threadname):strdup("")),
    tid_(~0UL),refcount_(1),isrunning_(false)
{
}

const char *
RACE_Thread::name()
{
    return name_;
}

RACE_Thread::~RACE_Thread()
{
  debug->print("RACE_Thread::~RACE_Thread() %s",name_);
  free((void *)name_);
}

void RACE_Thread::start()
{
  {
    ACE_GUARD(ACE_Thread_Mutex,guard,refcount_mutex_);
    refcount_++;
  }
  long flags=THR_JOINABLE | THR_NEW_LWP;// THR_DETACHED | THR_NEW_LWP;
      
  ACE_Thread::spawn(ACE_THR_FUNC(proc_to_run), this,
		    flags,0,&tid_);
}

void *RACE_Thread::proc_to_run(void *myself)
{
  RACE_Thread *w = (RACE_Thread *)myself;
  {
    ACE_GUARD_RETURN(ACE_Thread_Mutex,guard,w->mutex_,NULL);
    w->tid_=ACE_Thread::self();
    w->isrunning_=true;
    debug->print("RACE_Thread %s @TID is started , isrunning=%s\n",w->name(),(w->isrunning_ ? "TRUE":"FALSE"));
    w->state_changed_.signal();
  }
  w->run();
  {
    ACE_GUARD_RETURN(ACE_Thread_Mutex,guard,w->mutex_,NULL);
    w->isrunning_=false;
    w->state_changed_.signal();
    w->tid_=~0UL;
    debug->print("RACE_Thread %s @TID is stopped , isrunning=%s\n",w->name(),(w->isrunning_ ? "TRUE":"FALSE"));
  }
  w->release();
  return NULL;
}

bool RACE_Thread::is_running()
{
    ACE_GUARD_RETURN(ACE_Thread_Mutex,guard,mutex_,false);
    return isrunning_;
}

void RACE_Thread::waitforStatechanging()
{
    ACE_GUARD(ACE_Thread_Mutex,guard,mutex_);
    debug->print("Wait in @TID for Thread %s %lu State changing: %s",
		 name(),tid_,
		 isrunning_ ? "running -> stopped" : "stopped -> running"
	);
    state_changed_.wait();
}

void RACE_Thread::wait()
{
    if(tid_==ACE_Thread::self())
    {
	ERRLOG << RACE_LogMsg_Stream::ERROR << "RACE_Thread::wait: "
	       << "Can not wait on myself !!! " << E_EOM;
    }
    else if(tid_!=~0UL)
    {
	debug->print("Joining Thread %s %lu (mytid=@TID)\n",name(),tid_);
	ACE_hthread_t tid=tid_;
	ACE_Thread::join(tid_);
	debug->print("Thread %s %lu joined (%lu)\n",name(),tid,tid_);
    }
    else
    {
	debug->print("RACE_THREAD::wait: Thread %s isn't running",name());
    }
}

void RACE_Thread::release()
{
  bool instance_obsolete = false;
  {
    ACE_GUARD(ACE_Thread_Mutex,guard,refcount_mutex_);
    debug->print("RACE_Thread::release() %s",name());
    refcount_--;
    if(refcount_==0)
    {
      instance_obsolete = true;
    }
  }
  if(instance_obsolete)
  {
    delete this;
  }
}

int RACE_Thread::number_of_refs()
{
  ACE_GUARD_RETURN(ACE_Thread_Mutex,guard,refcount_mutex_,-1);
  return refcount_;
}

RACE_Thread *RACE_Thread::duplicate()
{
  ACE_GUARD_RETURN(ACE_Thread_Mutex,guard,refcount_mutex_,NULL);
  debug->print("RACE_Thread::duplicate");
  refcount_++;
  return this;
}

int
RACE_Thread::wait_for_all_threads (const ACE_Time_Value * /*timeout*/)
{
    //return ace_thread_manager.wait (timeout);
    return 0;
}
