#ifndef RACE_ERRLOG_LOGBUF_H
#define RACE_ERRLOG_LOGBUF_H
/**
 * after http://www.paulgrenyer.dyndns.org/articles/digging%20a%20ditch.pdf
 */
#include <streambuf> 
#include <vector>
#include <algorithm>

namespace RACE_ERRLOG
{

//##Documentation
//## Interface for the Log
//## The Logbuf writes the collected entry through this interface

template<class charT,class traits = std::char_traits<charT> >
class LogSink
{
public:
   virtual void log_write(const charT *c,std::streamsize l)=0;
};

//##Documentation
//## collects message from an ostream and writes it in a LogSink
template<class charT,class traits = std::char_traits<charT> >
class Logbuf : public std::basic_streambuf<charT,traits>
{
public:
  Logbuf():buffer_(),sink_(0) {}
  Logbuf(LogSink<charT,traits> *sink):buffer_(),sink_(sink) {}

  void setSink(LogSink<charT,traits> *sink) 
  {
      sink_=sink;
  }

  ~Logbuf()
  {
      sync();
  }

private:
  typedef typename std::basic_streambuf< charT, traits >::int_type int_type;
  typedef charT char_type;
  typedef std::vector< charT > buffer_type;
  buffer_type buffer_;
  LogSink<charT,traits> *sink_;

  //##Documentation
  //## optimization for multiple characters
  virtual std::streamsize xsputn( const char_type* s, std::streamsize num )
  {
      std::copy( s, s + num, std::back_inserter< buffer_type >( buffer_ ) );
      return num;
  }

  //##Documentation
  //## since we use not the buffer of basic_streambuf, each character
  //## is passed here
  virtual int_type overflow( int_type c )
  {
      if ( !traits::eq_int_type( c, traits::eof() ) )
      {
	  buffer_.push_back( c );
	  return traits::not_eof(c);
      }
      else
      {
	  return traits::eof();
      }
  }

  //##Documentation
  //## the flush or endl call this method
  //## here we trigger the writing in the Logsink
  virtual int sync()
  {
      if(!buffer_.empty() && sink_!=0)
      {
	  sink_->log_write(&buffer_[0],std::streamsize(buffer_.size()));
          buffer_.clear(); 
      }
      return 0;
  }
};

  //##Documentation
  //## The ostream must inferit from this to garantee the Logbuf
  //## is initialized first 
  template<class charT,class traits = std::char_traits<charT> >
  class logbuf_init
  {
    public:
      logbuf_init():buf_() {}
      logbuf_init(LogSink<charT,traits> *sink):buf_(sink) {}
      Logbuf<charT,traits> &buf() { return buf_; }
      Logbuf<charT,traits> *buf_ptr() { return &buf_; }
    private:
      Logbuf<charT,traits> buf_;
  };
} // RACE_ERRLOG
#endif
