#ifndef RACE_LOG_MSG_STREAM_H
#define RACE_LOG_MSG_STREAM_H
#include<ostream>
#include<sstream>
#include <cstdarg>
#include<ace/Synch.h>
#include<ace/Singleton.h>
#include"RACE_ERRLOG_Logbuf.h"
// Backends
#include"RACE_LogMsg_BackendMgr.h"
#include"RACE_LogMsg_Bstderr.h"
#include"RACE_LogMsg_Bsyslog.h"

#define RACE_MAXLOGMSGLEN 255

//##Documentation
//## An ostream in the Thread Specific Storage (no koordination is needing during assembling the message string)
//## With the help of the internaly used RACE_ERRLOG::Logbuf it can be used like a common ostream,
//## that means using operator << and std::endl and so on.
//## Further it has the API from output
class RACE_LogMsg_Stream : private virtual RACE_ERRLOG::logbuf_init<char>,public std::basic_ostream<char>,public RACE_ERRLOG::LogSink<char>
{
public:
    enum  Priority
    {
        EMERGENCY  = LOG_EMERG,   /// System is unusable.
        ALERT   = LOG_ALERT,	  /// Action must be taken immediately.
        CRIT    = LOG_CRIT,	  /// Critical condition.
        ERROR     = LOG_ERR,	  /// Error.
        WARNING = LOG_WARNING,	  /// Warning.
        NOTICE  = LOG_NOTICE,	  /// Normal, but important event.
        INFO    = LOG_INFO,    
        LDEBUG   = LOG_DEBUG	  /// Debugging message.
    };
    
    RACE_LogMsg_Stream();

    //##Documentation
    //## set the conditional data (see ACE_Log_Msg::setconditionald)
    //## and resets the ostream-data
    //## This is used in the ERRLOG-Macro. It returns the ostream for
    //## further use with the <<-Operators
    ostream& info(char *file,int line,int op_status,int errnum);

    //##Documentation
    //## from Interface RACE_ERRLOG::Logsink
    void log_write(const char *s,std::streamsize n);

    //##Documentation
    //## format like printf(3) and:
    //## @PROG : name of the executable
    //## @PID  : process id of the first prog
    //## @TID  : Calling thread id
    //## @FILE : source code file in witch the error occured
    //## @LINE : LineNumer where the error Occurd
    //## @ERRSTR or %m : strerror of the errnum
    //## @TIME : print timestamp in hour:minute:sec:usec format.
    //## implementation delegates to ACE_LogMsg::log
    ssize_t log(Priority priority, const ACE_TCHAR *format, ...);

    //##Documentation
    //## An alternative logging mechanism that makes it possible to
    //## integrate variable argument lists from other logging mechanisms
    //## into the RACE mechanism.
   ssize_t log (const ACE_TCHAR *format,
		Priority priority,
		va_list argp);

    //##Documentation
    //## Convenience API - writing a debugmessage
    void debug (char *msg, ...) __attribute__ ((format(printf, 2, 3)));

    //##Documentation
    //## Convenience API - writing a info-message
    void message (char *msg, ...) __attribute__ ((format(printf, 2, 3)));

    //##Documentation
    //## Convenience API - writing a warning-message
    void warn (char *msg, ...) __attribute__ ((format(printf, 2, 3)));

    //##Documentation
    //## Convenience API - writing a error-message
    void error (char *msg, ...) __attribute__ ((format(printf, 2, 3)));

    //##Documentation
    //## Convenience API - writing a notice-message	
    void report (char *msg, ...) __attribute__ ((format(printf, 2, 3)));

    //##Documentation
    //## Convenience API - writing a fatal error message and abort
    void fail (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
                                                                 
    //##Documentation
    //## Convenience API - writing a debug-message if validity is true
    void cond_debug (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    //##Documentation
    //## Convenience API - writing a notice-message if validity is true
    void cond_message (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    //##Documentation
    //## Convenience API - writing a warning-message if validity is true
    void cond_warn (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    //##Documentation
    //## Convenience API - writing a error-message if validity is true
    void cond_error (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    //##Documentation
    //## Convenience API - wring a fatal message and abort if validity is true
    void cond_fail (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    //##Documentation
    //## currently empty - intended for
    //## turning on printing a timestamp
    void time_output_on() {}
  
    //Accessoren
    void file(const char *filename);

    const char *file();
protected:
    RACE_LogMsg_BackendMgr *backend_;

    Priority m_defaultPriority;

    Priority m_currentPriority;

    friend std::ostream&  operator<<(std::ostream & os, RACE_LogMsg_Stream::Priority priority);

private:
    /// Status of operation (-1 means failure, >= 0 means success).
  int status_;
  /// Type of error that occurred (see <sys/errno.h>).
  int errnum_;
  
  /// Line number where the error occurred.
  int linenum_;
  
  /// File where the error occurred.
  char file_[MAXPATHLEN + 1];
                                                                                
  /// The log message, which resides in thread-specific storage.  Note
  /// that only the current log message is stored here -- it will be
  /// overwritten by the subsequent call to log().
  char msg_[RACE_MAXLOGMSGLEN + 1]; // Add one for NUL-terminator.
};


typedef ACE_TSS_Singleton<RACE_LogMsg_Stream,ACE_Recursive_Thread_Mutex> RACELOGMSGSTREAM;

#define ERRLOG RACELOGMSGSTREAM::instance()->info(__FILE__, __LINE__, 0 , 0) 

#define E_CLR ""
#define E_EOM std::endl

//##Documentation
//##
//## Set the default priority.
//##
//## @param priority to set as default..
std::ostream&  operator<<(std::ostream & os, ACE_Log_Priority priority);
#endif
