#include<string.h>
#include"time64.h"
#include<errorhandling/envdebug.h>
#include"Collector.h"
#include"KVDumpable.h"
#include"CollectorLoggerStrategy.h"
#include<RACE_DepotSwitchingStrategy.h>
#include<RACE_Logfile.h>
#include"KVInterpreter.h"
#include"KVFeedingSemantik.h"
#include"KVSnapshotSemantik.h"

static envdebug *debug = new envdebug("Collector");
Collector::Collector(const char *rumpfpfad,int genlog)
  :genlog_(genlog),snapshotdumpable_(0),
   snapshottrigger_("KV"),
   dirty_(false),
   informsnapshot_(0),
   refcount_(1)
{
  strncpy(rumpfpfad_,rumpfpfad,128);
  snapshotdumpable_=new KVSnapshotDumpable();
  RACE_MEMORY_FAIL(snapshotdumpable_==NULL,__FILE__,__LINE__,__FUNCTION__);
}

Collector::~Collector()
{
  snapshotdumpable_->release();
}

int Collector::open() throw(Databaseerror)
{
  RACE_LogSwitchingStrategy *lsstrat=0;
  if(genlog_)
  {
    logfile_=new RACE_Logfile();
    informsnapshot_=NULL;
  }
  else
  {
    KVInterpreter *interp      = new KVInterpreter();
    KVFeedingSemantik *itersem = new KVFeedingSemantik(this);
    logfile_=informsnapshot_   = new RACE_Depot(this,interp,itersem);
    informsnapshot_->SetSnapshotSemantik(new KVSnapshotSemantik(informsnapshot_->Reader(),interp,itersem));
    lsstrat =new RACE_DepotLogSwitchingStrategy();
    RACE_MEMORY_FAIL(lsstrat==NULL,__FILE__,__LINE__,__FUNCTION__);
  }
  RACE_MEMORY_FAIL(logfile_==NULL,__FILE__,__LINE__,__FUNCTION__);
  logfile_->Open(rumpfpfad_);
  RACE_LoggerStrategy *strat = new CollectorLoggerStrategy(rumpfpfad_,this);
  RACE_MEMORY_FAIL(strat==NULL,__FILE__,__LINE__,__FUNCTION__);
  logger_=new RACE_Logger(logfile_,strat,lsstrat);
  RACE_MEMORY_FAIL(logger_==NULL,__FILE__,__LINE__,__FUNCTION__);
  logger_->activate();
  if(informsnapshot_)
  {
    informsnapshot_->Recover();
  }
  return 0;
}

int Collector::close()
{
  Checkpoint();
  logger_->Cancel();
  logger_->wait();
  logger_->Release();
  logfile_->Close();
  return 0;
}

int Collector::fetch(Gdat &key,Gdat &val) throw(Databaseerror)
{
  return 0;
}

int Collector::store(Gdat &key,Gdat &value,Dmapflag flag) throw(Databaseerror)
{
  Timestamp_t now = timestamp_now();
  return store(key,now,value,flag);
}

int Collector::store(Gdat &key,Timestamp_t timestamp,Gdat &value,Dmapflag flag) throw(Databaseerror)
{
  int r=store_i(key,timestamp,value,flag);
  if(r==1 || genlog_) // new record or we are in logmode
  {
    KVDumpable *kvd = new KVDumpable(key,timestamp,value);
    RACE_COND_FAIL(kvd==NULL,"%s:%d: No Memory",__FILE__,__LINE__);
    kvd->SetConnection(this);
    IncrPending();
    logger_->Log(kvd);
  }
  return r;
}

int Collector::insert(KVDumpable *kvd)
{
  Gdat key,value;
  kvd->GetKey(key);
  kvd->GetValue(value);
  int r=store_i(key,kvd->GetTimestamp(),value,DMAP_REPLACE);
  if(r==1) // new record
  {
    KVDumpable *kvdi=dynamic_cast<KVDumpable *>(kvd->duplicate());
    kvdi->SetConnection(this);
    IncrPending();
    logger_->Log(kvdi);
  }
  return r;
}

int Collector::store_i(Gdat &key,Timestamp_t timestamp,Gdat &value,Dmapflag flag) throw(Databaseerror)
{
  KVTreeImpl::ENTRY *entry;
  int r;
  // Note: We make the insertion here since we store the value.
  // An alternative would be to store the Entry in the InformOffsetCallback
  RACE_GUARD_THROW_EX(ACE_Thread_Mutex,guard,mutex_,Databaseerror("Collector::store_i: Can't accquire Mutex"));
  TreeEntry_t rec(key,timestamp,value);
  TreeEntry_t newrec(rec); // Copy is needed because trybind overwrides
  r=kvtree_.trybind(key,rec,entry);
  if(r<0) throw  Databaseerror("ACE_RBTree: bind failed");
  if(r==0)
  { /* new entry */
    debug->print("Collector::store_i: New Value >%*s<",key.get_len(),(const char *)key.get_const_data());
    dirty_=true;
    return 1;
  }
  if(r==1 && newrec.timestamp > rec.timestamp)
  {
    debug->print("Collector::store_i: Overwrite Value >%*s<",key.get_len(),(const char *)key.get_const_data());
    entry->item() = newrec;
    dirty_=true;
  }
  else
  {
    debug->print("Collector::store_i:Ignore Value >%*s<",key.get_len(),(const char *)key.get_const_data());
    return 0;
  }
  return r;
}

void Collector::RewriteAllAndClear()
{
  debug->print("RewriteAllAndClear()");
  snapshotdumpable_->release();
  snapshotdumpable_=new KVSnapshotDumpable();
  RACE_MEMORY_FAIL(snapshotdumpable_==NULL,__FILE__,__LINE__,__FUNCTION__);
  for(KVTreeImpl::ITERATOR i(kvtree_);!i.done();i.advance())
  {
    (*i).item().entry=NULL;
    KVDumpable *kvd = new KVDumpable((*i).item().key,(*i).item().timestamp,(*i).item().value);
    RACE_COND_FAIL(kvd==NULL,"%s:%d: No Memory",__FILE__,__LINE__);
    kvd->SetConnection(this);
    IncrPending();
    logger_->Log(kvd);
  }
}

void Collector::asciidump(std::ostream& str)
{
  for(KVTreeImpl::ITERATOR i(kvtree_);!i.done();i.advance())
  {
    Logfile::KeyRecord* entry=(*i).item().entry;
    str << (*i).item().key << "   " << (*i).item().value;
    if(entry)
    {
      str << " at " << entry->offset.part << ':' << entry->offset.offset;
    }
    str << std::endl;
  }
}

void Collector::Checkpoint(time_t timestamp)
{
  debug->print("Collector::Checkpoint(%ld):",timestamp);
  if(dirty_ || GetPendingTotal()>0)
  {
    snapshottrigger_.PrepareSnapshot(this, timestamp);
    dirty_=false;
    logger_->Flush();
  }
  else
  {
    debug->print("Collector::Checkpoint: No Checkpoint neccessary at %ld",timestamp);
  }

}

void  Collector::Checkpoint()
{
  time_t t;
  time(&t);
  Checkpoint(t);
}

void Collector::SetOffset(RACE_LogfileOffset offset,time_t timestamp)
{
  debug->print("Snapshot %ld at %d:%d",timestamp,offset.part,offset.offset);
  logfile_->WriteIndexEntry(timestamp,offset);
  DecrSnapshotPending();
}

void Collector::InformSnapshotWritten()
{
  if(informsnapshot_)
  {
    //In the Depository-Case delete the old file
    informsnapshot_->InformSnapshot();
  }
  // if our snapshot is on the disk, inform the Stategy.
  // it may decide to switch the logfile here.
  logger_->GetSwitchingStrategy()->InfoSnapshotWritten();  
}

void Collector::InformDiscard(time_t timestamp)
{
  /* Ignore */
  snapshottrigger_.InformInsert();
  DecrPending(); // Our Instance may delete here
}

void Collector::KeyValueInsertCallback(const Gdat& key,RACE_LogfileOffset offset)
{
  debug->print("KeyValueInsertCallback(%*s at %d:%d)",key.get_len(),(const char *)key.get_const_data(),offset.part,offset.offset);
  int r=0;
  KVTreeImpl::ENTRY *node=NULL;
  { ACE_GUARD(ACE_Thread_Mutex,guard,mutex_);
    r=kvtree_.find(key,node);
    if(r==0)
    {
      if(node->item().entry == NULL)
      {
        // New Entry
        debug->print("New Key: %*s",key.get_len(),(const char *)key.get_const_data());
        Logfile::KeyRecord *keyrec=snapshotdumpable_->append(key);
        node->item().entry = keyrec;
        keyrec->offset = offset;
      }
      else
      {
        debug->print("Key %*s found",key.get_len(),(const char *)key.get_const_data());
        node->item().entry->offset = offset;
      }
      dirty_=true; // We have written in the KVSnapshotDumpable
    }
    else
    {
      //Happens, if a key is deleted
      //and the callback comes after the timeout time
      debug->print("Collector::KeyValueInsertCallback: key not found");
    }
  }//ACE_Guard
  snapshottrigger_.InformInsert();
  DecrPending(); // Our Instance may delete here
}

void Collector::KeyValueDiscardCallback(const Gdat& key)
{
  debug->print("KeyValueDiscardCallback(%*s)",key.get_len(),(const char *)key.get_const_data());
  snapshottrigger_.InformInsert();
  DecrPending(); // Our Instance may delete here
}

void Collector::IncrPending()
{
  pending_++;
}

void Collector::DecrPending()
{
  bool suicide;
  mutex_.acquire();
  pending_--;
  suicide=!CheckRightToExist();
  mutex_.release();
  if(suicide)
  {
    debug->print(" Collector::DecrPending: deleting ");
    delete this;
  }
}

void Collector::IncrSnapshotPending()
{
  pendingsnapshots_++;
}

void Collector::DecrSnapshotPending()
{
  bool suicide;
  mutex_.acquire();
  pendingsnapshots_--;
  suicide=!CheckRightToExist();
  mutex_.release();
  if(suicide)
  {
    debug->print(" Collector::DecrSnapshotPending: deleting ");
    delete this;
  }
}

void Collector::IncrWorking()
{
  working_++;
  RACE_COND_FAIL(working_>1,"%s:%d: Working > 1 ",__FILE__,__LINE__);
}

void Collector::DecrWorking()
{
  bool suicide;
  mutex_.acquire();
  working_--;
  suicide=!CheckRightToExist();
  mutex_.release();
  if(suicide)
  {
    debug->print(" Collector::DecrWorking: deleting ");
    delete this;
  }
}

void Collector::release()
{
  bool suicide;
  mutex_.acquire();
  refcount_--;
  suicide=!CheckRightToExist();
  mutex_.release();
  if(suicide)
  {
    debug->print(" Collector::release: deleting ");
    delete this;
  }
}

bool Collector::CheckRightToExist()
{
  bool r=(refcount_>0 || pending_>0 || pendingsnapshots_>0||working_>0);
  debug->print("CheckRightToExist=%s:refcount=%ld,pending_=%ld,pendingsnapshots=%ld,working_=%ld",(r? "true":"false"),refcount_.value(),pending_.value(),pendingsnapshots_.value(),working_.value());
  return r;
}

int Collector::GetPendingTotal()
{
  return int(pending_.value());
}

RACE_SnapshotQueue::Entry *Collector::GenNewEntry(time_t timestamp,int pending)
{
  KVSSnapshotEntry *ret = new KVSSnapshotEntry(timestamp,pending,this,snapshotdumpable_,logger_);
  RACE_COND_FAIL(ret==NULL,"%s:%d: No Memory",__FILE__,__LINE__);
  return ret;
}

void KVSSnapshotEntry::ConnectionLog()
{
  KVSnapshotDumpable *copy = ACE_dynamic_cast(KVSnapshotDumpable *,snapshot_->clone());
  RACE_COND_FAIL(copy==NULL,"KVSSnapshotEntry::ConnectionLog(): narrowing copy failed");
  copy->SetLogSnapshotConnection(pcd_,GetTimestamp());
  pcd_->IncrSnapshotPending();
  logger_->Log(copy);
  logger_->Flush(); // The logfile may switch ...
}

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)
template class ACE_RB_Tree<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>;
template class ACE_RB_Tree_Node<Gdat_t,Collector::TreeEntry_t>;
template class ACE_RB_Tree_Iterator_Base<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>;
template class ACE_RB_Tree_Iterator<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>;
template class ACE_RB_Tree_Reverse_Iterator<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>;
#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)
#pragma instantiate ACE_RB_Tree<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>
#pragma instantiate ACE_RB_Tree_Node<Gdat_t,Collector::TreeEntry_t>
#pragma instantiate ACE_RB_Tree_Iterator_Base<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>
#pragma instantiate ACE_RB_Tree_Iterator<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>
#pragma instantiate ACE_RB_Reverse_Tree_Iterator<Gdat_t,Collector::TreeEntry_t,GdatCmp,ACE_Null_Mutex>
#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */

