#ifndef RACE_LOG_MSG_STREAM_H
#define RACE_LOG_MSG_STREAM_H
#include<ostream>
#include<sstream>
#include <cstdarg>
#include<ace/Log_Msg.h>
#include<ace/Synch.h>
#include<ace/Singleton.h>

class RACE_Log_Msg_Stream : public std::ostream
{
public:
    //##Documentation
    //##
    //## Control - control the logging stream
    //##
    //## eom:
    //## If you use the stream output operator<<, you have to mark
    //## the end of message (eom) to send the message to the syslogd.
    //## This allows to concatenate several strings to one message:
    //## syslog << "This is a " << "log message." << Syslog::eom;
    //## will produce: <identifier>[pid]: This is a log message.
    //##
    //## clr:
    //## Clear the message buffer.
    typedef enum
      {
        eom,
        clr
      }Control;
  
    RACE_Log_Msg_Stream();
    ostream& info(char *file,int line,int op_status,int errnum);
    //##Documentation
    //## see ACE_Log_Msg::log 
    //## implementation delegates to ACE_Log_Msg::log
    ssize_t log(ACE_Log_Priority priority, const ACE_TCHAR *format, ...);

    //##Documentation
    //## An alternative logging mechanism that makes it possible to
    //## integrate variable argument lists from other logging mechanisms
    //## into the RACE mechanism.
   ssize_t log (const ACE_TCHAR *format,
               ACE_Log_Priority priority,
               va_list argp);

    void debug (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
    void message (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
    void warn (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
    void error (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
    void report (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
    void fail (char *msg, ...) __attribute__ ((format(printf, 2, 3)));
                                                                                
    void cond_debug (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));
    void cond_message (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));
    void cond_warn (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));
    void cond_error (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));
    void cond_fail (bool validity, char *msg, ...) __attribute__ ((format(printf,3,4)));

    void time_output_on() {}
    //##Documentation
    //## see ACE_Log_Msg::open
    //## implementation delegates to ACE_Log_Msg::log
    int open (const ACE_TCHAR *prog_name,
	      u_long flags = ACE_Log_Msg::STDERR,
	      const ACE_TCHAR *logger_key=0);
protected:
    ACE_Log_Msg *acelog_;
    ACE_Log_Priority m_defaultPriority;
    ACE_Log_Priority m_currentPriority;

    //##Documentation
    //## Message storage for stream operator<<.
    std::ostringstream m_strbuf;

    friend std::ostream&  operator<<(std::ostream & os, RACE_Log_Msg_Stream::Control c);
    friend std::ostream&  operator<<(std::ostream & os, ACE_Log_Priority priority);
};


typedef ACE_TSS_Singleton<RACE_Log_Msg_Stream,ACE_Recursive_Thread_Mutex> RACELOGMSGSTREAM;

#define ERRLOG RACELOGMSGSTREAM::instance()->info(__FILE__, __LINE__, 0 , ACE_Log_Msg::last_error_adapter()) 

#define E_CLR RACE_Log_Msg_Stream::clr
#define E_EOM RACE_Log_Msg_Stream::eom;
//##Documentation
//##
//## Apply a control function to the message buffer.
//##
//## For available control functions @see Control
//##
//## @param control define control function
std::ostream & operator<<(std::ostream & os, RACE_Log_Msg_Stream::Control control);

//##Documentation
//##
//## Set the default priority.
//##
//## @param priority to set as default..
std::ostream&  operator<<(std::ostream & os, ACE_Log_Priority priority);
#endif
